@@ -2,25 +2,23 @@ use crate::{
     bank::{Builtin, Builtins},
     system_instruction_processor,
 };
-use solana_sdk::{
+use safecoin_sdk::{
     instruction::InstructionError,
-    keyed_account::KeyedAccount,
     process_instruction::{stable_log, InvokeContext, ProcessInstructionWithContext},
     pubkey::Pubkey,
-    system_program,
+    stake, system_program,
 };
 
 fn process_instruction_with_program_logging(
     process_instruction: ProcessInstructionWithContext,
     program_id: &Pubkey,
-    keyed_accounts: &[KeyedAccount],
     instruction_data: &[u8],
     invoke_context: &mut dyn InvokeContext,
 ) -> Result<(), InstructionError> {
     let logger = invoke_context.get_logger();
     stable_log::program_invoke(&logger, program_id, invoke_context.invoke_depth());
 
-    let result = process_instruction(program_id, keyed_accounts, instruction_data, invoke_context);
+    let result = process_instruction(program_id, instruction_data, invoke_context);
 
     match &result {
         Ok(()) => stable_log::program_success(&logger, program_id),
@@ -31,14 +29,10 @@ fn process_instruction_with_program_logging(
 
 macro_rules! with_program_logging {
     ($process_instruction:expr) => {
-        |program_id: &Pubkey,
-         keyed_accounts: &[KeyedAccount],
-         instruction_data: &[u8],
-         invoke_context: &mut dyn InvokeContext| {
+        |program_id: &Pubkey, instruction_data: &[u8], invoke_context: &mut dyn InvokeContext| {
             process_instruction_with_program_logging(
                 $process_instruction,
                 program_id,
-                keyed_accounts,
                 instruction_data,
                 invoke_context,
             )
@@ -61,7 +55,7 @@ fn genesis_builtins() -> Vec<Builtin> {
         ),
         Builtin::new(
             "stake_program",
-            solana_stake_program::id(),
+            stake::program::id(),
             with_program_logging!(solana_stake_program::stake_instruction::process_instruction),
         ),
         Builtin::new(
@@ -71,7 +65,7 @@ fn genesis_builtins() -> Vec<Builtin> {
         ),
         Builtin::new(
             "secp256k1_program",
-            solana_sdk::secp256k1_program::id(),
+            safecoin_sdk::secp256k1_program::id(),
             solana_secp256k1_program::process_instruction,
         ),
     ]
