@@ -1,15 +1,23 @@
 //! Stakes serve as a cache of stake and vote accounts to derive
 //! node stakes
-use crate::vote_account::{ArcVoteAccount, VoteAccounts};
-use solana_sdk::{
+use crate::vote_account::{ArcVoteAccount, VoteAccounts, VoteAccountsHashMap};
+use rayon::{
+    iter::{IntoParallelRefIterator, ParallelIterator},
+    ThreadPool,
+};
+use safecoin_sdk::{
     account::{AccountSharedData, ReadableAccount},
     clock::Epoch,
     pubkey::Pubkey,
-    sysvar::stake_history::StakeHistory,
+    stake::{
+        self,
+        state::{Delegation, StakeActivationStatus, StakeState},
+    },
+    stake_history::StakeHistory,
 };
-use solana_stake_program::stake_state::{new_stake_history_entry, Delegation, StakeState};
+use solana_stake_program::stake_state;
 use solana_vote_program::vote_state::VoteState;
-use std::{borrow::Borrow, collections::HashMap};
+use std::collections::HashMap;
 
 #[derive(Default, Clone, PartialEq, Debug, Deserialize, Serialize, AbiExample)]
 pub struct Stakes {
@@ -33,7 +41,9 @@ impl Stakes {
     pub fn history(&self) -> &StakeHistory {
         &self.stake_history
     }
-    pub fn clone_with_epoch(&self, next_epoch: Epoch, fix_stake_deactivate: bool) -> Self {
+
+    #[deprecated(note = "remove after optimize_epoch_boundary_updates feature is active")]
+    pub fn clone_with_epoch(&self, next_epoch: Epoch) -> Self {
         let prev_epoch = self.epoch;
         if prev_epoch == next_epoch {
             self.clone()
@@ -42,13 +52,12 @@ impl Stakes {
             let mut stake_history_upto_prev_epoch = self.stake_history.clone();
             stake_history_upto_prev_epoch.add(
                 prev_epoch,
-                new_stake_history_entry(
+                stake_state::new_stake_history_entry(
                     prev_epoch,
                     self.stake_delegations
                         .iter()
                         .map(|(_pubkey, stake_delegation)| stake_delegation),
                     Some(&self.stake_history),
-                    fix_stake_deactivate,
                 ),
             );
 
@@ -61,7 +70,6 @@ impl Stakes {
                         pubkey,
                         next_epoch,
                         Some(&stake_history_upto_prev_epoch),
-                        fix_stake_deactivate,
                     );
                     (*pubkey, (stake, account.clone()))
                 })
@@ -77,19 +85,85 @@ impl Stakes {
         }
     }
 
+    pub fn activate_epoch(&mut self, next_epoch: Epoch, thread_pool: &ThreadPool) {
+        let prev_epoch = self.epoch;
+        self.epoch = next_epoch;
+
+        thread_pool.install(|| {
+            let stake_delegations = &self.stake_delegations;
+            let stake_history = &mut self.stake_history;
+            let vote_accounts: &VoteAccountsHashMap = self.vote_accounts.as_ref();
+
+            // construct map of vote pubkey -> list of stake delegations
+            let vote_delegations: HashMap<Pubkey, Vec<&Delegation>> = {
+                let mut vote_delegations = HashMap::with_capacity(vote_accounts.len());
+                stake_delegations
+                    .iter()
+                    .for_each(|(_stake_pubkey, delegation)| {
+                        let vote_pubkey = &delegation.voter_pubkey;
+                        vote_delegations
+                            .entry(*vote_pubkey)
+                            .and_modify(|delegations: &mut Vec<_>| delegations.push(delegation))
+                            .or_insert_with(|| vec![delegation]);
+                    });
+                vote_delegations
+            };
+
+            // wrap up the prev epoch by adding new stake history entry for the prev epoch
+            {
+                let stake_history_entry = vote_delegations
+                    .par_iter()
+                    .map(|(_vote_pubkey, delegations)| {
+                        delegations
+                            .par_iter()
+                            .map(|delegation| {
+                                delegation.stake_activating_and_deactivating(
+                                    prev_epoch,
+                                    Some(stake_history),
+                                )
+                            })
+                            .reduce(StakeActivationStatus::default, |a, b| a + b)
+                    })
+                    .reduce(StakeActivationStatus::default, |a, b| a + b);
+
+                stake_history.add(prev_epoch, stake_history_entry);
+            }
+
+            // refresh the stake distribution of vote accounts for the next epoch, using new stake history
+            let vote_accounts_for_next_epoch: VoteAccountsHashMap = vote_accounts
+                .par_iter()
+                .map(|(vote_pubkey, (_stake, vote_account))| {
+                    let delegated_stake = vote_delegations
+                        .get(vote_pubkey)
+                        .map(|delegations| {
+                            delegations
+                                .par_iter()
+                                .map(|delegation| delegation.stake(next_epoch, Some(stake_history)))
+                                .sum()
+                        })
+                        .unwrap_or_default();
+
+                    (*vote_pubkey, (delegated_stake, vote_account.clone()))
+                })
+                .collect();
+
+            // overwrite vote accounts so that staked nodes singleton is reset
+            self.vote_accounts = VoteAccounts::from(vote_accounts_for_next_epoch);
+        });
+    }
+
     // sum the stakes that point to the given voter_pubkey
     fn calculate_stake(
         &self,
         voter_pubkey: &Pubkey,
         epoch: Epoch,
         stake_history: Option<&StakeHistory>,
-        fix_stake_deactivate: bool,
     ) -> u64 {
         self.stake_delegations
             .iter()
             .map(|(_, stake_delegation)| {
                 if &stake_delegation.voter_pubkey == voter_pubkey {
-                    stake_delegation.stake(epoch, stake_history, fix_stake_deactivate)
+                    stake_delegation.stake(epoch, stake_history)
                 } else {
                     0
                 }
@@ -110,8 +184,8 @@ impl Stakes {
     }
 
     pub fn is_stake(account: &AccountSharedData) -> bool {
-        solana_vote_program::check_id(&account.owner)
-            || solana_stake_program::check_id(&account.owner)
+        solana_vote_program::check_id(account.owner())
+            || stake::program::check_id(account.owner())
                 && account.data().len() >= std::mem::size_of::<StakeState>()
     }
 
@@ -119,28 +193,21 @@ impl Stakes {
         &mut self,
         pubkey: &Pubkey,
         account: &AccountSharedData,
-        fix_stake_deactivate: bool,
         check_vote_init: bool,
+        remove_delegation_on_inactive: bool,
     ) -> Option<ArcVoteAccount> {
-        if solana_vote_program::check_id(&account.owner) {
+        if solana_vote_program::check_id(account.owner()) {
             // unconditionally remove existing at first; there is no dependent calculated state for
             // votes, not like stakes (stake codepath maintains calculated stake value grouped by
             // delegated vote pubkey)
             let old = self.vote_accounts.remove(pubkey);
             // when account is removed (lamports == 0 or data uninitialized), don't read so that
             // given `pubkey` can be used for any owner in the future, while not affecting Stakes.
-            if account.lamports != 0
-                && !(check_vote_init && VoteState::is_uninitialized_no_deser(&account.data()))
+            if account.lamports() != 0
+                && !(check_vote_init && VoteState::is_uninitialized_no_deser(account.data()))
             {
                 let stake = old.as_ref().map_or_else(
-                    || {
-                        self.calculate_stake(
-                            pubkey,
-                            self.epoch,
-                            Some(&self.stake_history),
-                            fix_stake_deactivate,
-                        )
-                    },
+                    || self.calculate_stake(pubkey, self.epoch, Some(&self.stake_history)),
                     |v| v.0,
                 );
 
@@ -148,26 +215,22 @@ impl Stakes {
                     .insert(*pubkey, (stake, ArcVoteAccount::from(account.clone())));
             }
             old.map(|(_, account)| account)
-        } else if solana_stake_program::check_id(&account.owner) {
+        } else if stake::program::check_id(account.owner()) {
             //  old_stake is stake lamports and voter_pubkey from the pre-store() version
             let old_stake = self.stake_delegations.get(pubkey).map(|delegation| {
                 (
                     delegation.voter_pubkey,
-                    delegation.stake(self.epoch, Some(&self.stake_history), fix_stake_deactivate),
+                    delegation.stake(self.epoch, Some(&self.stake_history)),
                 )
             });
 
-            let delegation = StakeState::delegation_from(account);
+            let delegation = stake_state::delegation_from(account);
 
             let stake = delegation.map(|delegation| {
                 (
                     delegation.voter_pubkey,
-                    if account.lamports != 0 {
-                        delegation.stake(
-                            self.epoch,
-                            Some(&self.stake_history),
-                            fix_stake_deactivate,
-                        )
+                    if account.lamports() != 0 {
+                        delegation.stake(self.epoch, Some(&self.stake_history))
                     } else {
                         // when account is removed (lamports == 0), this special `else` clause ensures
                         // resetting cached stake value below, even if the account happens to be
@@ -187,7 +250,13 @@ impl Stakes {
                 }
             }
 
-            if account.lamports == 0 {
+            let remove_delegation = if remove_delegation_on_inactive {
+                delegation.is_none()
+            } else {
+                account.lamports() == 0
+            };
+
+            if remove_delegation {
                 // when account is removed (lamports == 0), remove it from Stakes as well
                 // so that given `pubkey` can be used for any owner in the future, while not
                 // affecting Stakes.
@@ -205,8 +274,8 @@ impl Stakes {
         }
     }
 
-    pub fn vote_accounts(&self) -> &HashMap<Pubkey, (u64, ArcVoteAccount)> {
-        self.vote_accounts.borrow()
+    pub fn vote_accounts(&self) -> &VoteAccountsHashMap {
+        self.vote_accounts.as_ref()
     }
 
     pub fn stake_delegations(&self) -> &HashMap<Pubkey, Delegation> {
@@ -230,7 +299,8 @@ impl Stakes {
 #[cfg(test)]
 pub mod tests {
     use super::*;
-    use solana_sdk::{pubkey::Pubkey, rent::Rent};
+    use rayon::ThreadPoolBuilder;
+    use safecoin_sdk::{account::WritableAccount, pubkey::Pubkey, rent::Rent};
     use solana_stake_program::stake_state;
     use solana_vote_program::vote_state::{self, VoteState, VoteStateVersions};
 
@@ -238,9 +308,9 @@ pub mod tests {
     pub fn create_staked_node_accounts(
         stake: u64,
     ) -> ((Pubkey, AccountSharedData), (Pubkey, AccountSharedData)) {
-        let vote_pubkey = solana_sdk::pubkey::new_rand();
+        let vote_pubkey = safecoin_sdk::pubkey::new_rand();
         let vote_account =
-            vote_state::create_account(&vote_pubkey, &solana_sdk::pubkey::new_rand(), 0, 1);
+            vote_state::create_account(&vote_pubkey, &safecoin_sdk::pubkey::new_rand(), 0, 1);
         (
             (vote_pubkey, vote_account),
             create_stake_account(stake, &vote_pubkey),
@@ -249,13 +319,13 @@ pub mod tests {
 
     //   add stake to a vote_pubkey                               (   stake    )
     pub fn create_stake_account(stake: u64, vote_pubkey: &Pubkey) -> (Pubkey, AccountSharedData) {
-        let stake_pubkey = solana_sdk::pubkey::new_rand();
+        let stake_pubkey = safecoin_sdk::pubkey::new_rand();
         (
             stake_pubkey,
             stake_state::create_account(
                 &stake_pubkey,
-                &vote_pubkey,
-                &vote_state::create_account(&vote_pubkey, &solana_sdk::pubkey::new_rand(), 0, 1),
+                vote_pubkey,
+                &vote_state::create_account(vote_pubkey, &safecoin_sdk::pubkey::new_rand(), 0, 1),
                 &Rent::free(),
                 stake,
             ),
@@ -266,9 +336,9 @@ pub mod tests {
         stake: u64,
         epoch: Epoch,
     ) -> ((Pubkey, AccountSharedData), (Pubkey, AccountSharedData)) {
-        let vote_pubkey = solana_sdk::pubkey::new_rand();
+        let vote_pubkey = safecoin_sdk::pubkey::new_rand();
         let vote_account =
-            vote_state::create_account(&vote_pubkey, &solana_sdk::pubkey::new_rand(), 0, 1);
+            vote_state::create_account(&vote_pubkey, &safecoin_sdk::pubkey::new_rand(), 0, 1);
         (
             (vote_pubkey, vote_account),
             create_warming_stake_account(stake, epoch, &vote_pubkey),
@@ -281,13 +351,13 @@ pub mod tests {
         epoch: Epoch,
         vote_pubkey: &Pubkey,
     ) -> (Pubkey, AccountSharedData) {
-        let stake_pubkey = solana_sdk::pubkey::new_rand();
+        let stake_pubkey = safecoin_sdk::pubkey::new_rand();
         (
             stake_pubkey,
             stake_state::create_account_with_activation_epoch(
                 &stake_pubkey,
-                &vote_pubkey,
-                &vote_state::create_account(&vote_pubkey, &solana_sdk::pubkey::new_rand(), 0, 1),
+                vote_pubkey,
+                &vote_state::create_account(vote_pubkey, &safecoin_sdk::pubkey::new_rand(), 0, 1),
                 &Rent::free(),
                 stake,
                 epoch,
@@ -308,41 +378,41 @@ pub mod tests {
 
             stakes.store(&vote_pubkey, &vote_account, true, true);
             stakes.store(&stake_pubkey, &stake_account, true, true);
-            let stake = StakeState::stake_from(&stake_account).unwrap();
+            let stake = stake_state::stake_from(&stake_account).unwrap();
             {
                 let vote_accounts = stakes.vote_accounts();
                 assert!(vote_accounts.get(&vote_pubkey).is_some());
                 assert_eq!(
                     vote_accounts.get(&vote_pubkey).unwrap().0,
-                    stake.stake(i, None, true)
+                    stake.stake(i, None)
                 );
             }
 
-            stake_account.lamports = 42;
+            stake_account.set_lamports(42);
             stakes.store(&stake_pubkey, &stake_account, true, true);
             {
                 let vote_accounts = stakes.vote_accounts();
                 assert!(vote_accounts.get(&vote_pubkey).is_some());
                 assert_eq!(
                     vote_accounts.get(&vote_pubkey).unwrap().0,
-                    stake.stake(i, None, true)
+                    stake.stake(i, None)
                 ); // stays old stake, because only 10 is activated
             }
 
             // activate more
             let (_stake_pubkey, mut stake_account) = create_stake_account(42, &vote_pubkey);
             stakes.store(&stake_pubkey, &stake_account, true, true);
-            let stake = StakeState::stake_from(&stake_account).unwrap();
+            let stake = stake_state::stake_from(&stake_account).unwrap();
             {
                 let vote_accounts = stakes.vote_accounts();
                 assert!(vote_accounts.get(&vote_pubkey).is_some());
                 assert_eq!(
                     vote_accounts.get(&vote_pubkey).unwrap().0,
-                    stake.stake(i, None, true)
+                    stake.stake(i, None)
                 ); // now stake of 42 is activated
             }
 
-            stake_account.lamports = 0;
+            stake_account.set_lamports(0);
             stakes.store(&stake_pubkey, &stake_account, true, true);
             {
                 let vote_accounts = stakes.vote_accounts();
@@ -394,7 +464,7 @@ pub mod tests {
             assert_eq!(vote_accounts.get(&vote_pubkey).unwrap().0, 10);
         }
 
-        vote_account.lamports = 0;
+        vote_account.set_lamports(0);
         stakes.store(&vote_pubkey, &vote_account, true, true);
 
         {
@@ -402,7 +472,7 @@ pub mod tests {
             assert!(vote_accounts.get(&vote_pubkey).is_none());
         }
 
-        vote_account.lamports = 1;
+        vote_account.set_lamports(1);
         stakes.store(&vote_pubkey, &vote_account, true, true);
 
         {
@@ -413,7 +483,9 @@ pub mod tests {
 
         // Vote account too big
         let cache_data = vote_account.data().to_vec();
-        vote_account.data.push(0);
+        let mut pushed = vote_account.data().to_vec();
+        pushed.push(0);
+        vote_account.set_data(pushed);
         stakes.store(&vote_pubkey, &vote_account, true, true);
 
         {
@@ -461,14 +533,14 @@ pub mod tests {
         // delegates to vote_pubkey
         stakes.store(&stake_pubkey, &stake_account, true, true);
 
-        let stake = StakeState::stake_from(&stake_account).unwrap();
+        let stake = stake_state::stake_from(&stake_account).unwrap();
 
         {
             let vote_accounts = stakes.vote_accounts();
             assert!(vote_accounts.get(&vote_pubkey).is_some());
             assert_eq!(
                 vote_accounts.get(&vote_pubkey).unwrap().0,
-                stake.stake(stakes.epoch, Some(&stakes.stake_history), true)
+                stake.stake(stakes.epoch, Some(&stakes.stake_history))
             );
             assert!(vote_accounts.get(&vote_pubkey2).is_some());
             assert_eq!(vote_accounts.get(&vote_pubkey2).unwrap().0, 0);
@@ -484,7 +556,7 @@ pub mod tests {
             assert!(vote_accounts.get(&vote_pubkey2).is_some());
             assert_eq!(
                 vote_accounts.get(&vote_pubkey2).unwrap().0,
-                stake.stake(stakes.epoch, Some(&stakes.stake_history), true)
+                stake.stake(stakes.epoch, Some(&stakes.stake_history))
             );
         }
     }
@@ -512,6 +584,7 @@ pub mod tests {
             assert_eq!(vote_accounts.get(&vote_pubkey).unwrap().0, 20);
         }
     }
+
     #[test]
     fn test_clone_with_epoch() {
         let mut stakes = Stakes::default();
@@ -521,21 +594,51 @@ pub mod tests {
 
         stakes.store(&vote_pubkey, &vote_account, true, true);
         stakes.store(&stake_pubkey, &stake_account, true, true);
-        let stake = StakeState::stake_from(&stake_account).unwrap();
+        let stake = stake_state::stake_from(&stake_account).unwrap();
 
         {
             let vote_accounts = stakes.vote_accounts();
             assert_eq!(
                 vote_accounts.get(&vote_pubkey).unwrap().0,
-                stake.stake(stakes.epoch, Some(&stakes.stake_history), true)
+                stake.stake(stakes.epoch, Some(&stakes.stake_history))
             );
         }
-        let stakes = stakes.clone_with_epoch(3, true);
+        #[allow(deprecated)]
+        let stakes = stakes.clone_with_epoch(3);
         {
             let vote_accounts = stakes.vote_accounts();
             assert_eq!(
                 vote_accounts.get(&vote_pubkey).unwrap().0,
-                stake.stake(stakes.epoch, Some(&stakes.stake_history), true)
+                stake.stake(stakes.epoch, Some(&stakes.stake_history))
+            );
+        }
+    }
+
+    #[test]
+    fn test_activate_epoch() {
+        let mut stakes = Stakes::default();
+
+        let ((vote_pubkey, vote_account), (stake_pubkey, stake_account)) =
+            create_staked_node_accounts(10);
+
+        stakes.store(&vote_pubkey, &vote_account, true, true);
+        stakes.store(&stake_pubkey, &stake_account, true, true);
+        let stake = stake_state::stake_from(&stake_account).unwrap();
+
+        {
+            let vote_accounts = stakes.vote_accounts();
+            assert_eq!(
+                vote_accounts.get(&vote_pubkey).unwrap().0,
+                stake.stake(stakes.epoch, Some(&stakes.stake_history))
+            );
+        }
+        let thread_pool = ThreadPoolBuilder::new().num_threads(1).build().unwrap();
+        stakes.activate_epoch(3, &thread_pool);
+        {
+            let vote_accounts = stakes.vote_accounts();
+            assert_eq!(
+                vote_accounts.get(&vote_pubkey).unwrap().0,
+                stake.stake(stakes.epoch, Some(&stakes.stake_history))
             );
         }
     }
@@ -562,7 +665,7 @@ pub mod tests {
         // not a stake account, and whacks above entry
         stakes.store(
             &stake_pubkey,
-            &AccountSharedData::new(1, 0, &solana_stake_program::id()),
+            &AccountSharedData::new(1, 0, &stake::program::id()),
             true,
             true,
         );
@@ -600,8 +703,9 @@ pub mod tests {
         assert_eq!(stakes.vote_balance_and_staked(), 11);
         assert_eq!(stakes.vote_balance_and_warmed_staked(), 1);
 
+        let thread_pool = ThreadPoolBuilder::new().num_threads(1).build().unwrap();
         for (epoch, expected_warmed_stake) in ((genesis_epoch + 1)..=3).zip(&[2, 3, 4]) {
-            stakes = stakes.clone_with_epoch(epoch, true);
+            stakes.activate_epoch(epoch, &thread_pool);
             // vote_balance_and_staked() always remain to return same lamports
             // while vote_balance_and_warmed_staked() gradually increases
             assert_eq!(stakes.vote_balance_and_staked(), 11);
