@@ -1,5 +1,5 @@
 use crossbeam_channel::{Receiver, Sender};
-use solana_sdk::{hash::Hash, pubkey::Pubkey};
+use safecoin_sdk::{hash::Hash, pubkey::Pubkey};
 use solana_vote_program::vote_state::Vote;
 
 pub type ReplayedVote = (Pubkey, Vote, Option<Hash>);
